"""REST helpers for Muffin Framework."""
import ujson
from muffin import HTTPNotFound, HTTPBadRequest, HTTPForbidden


# Package information
# ===================

__version__ = "1.2.1"
__project__ = "muffin-rest"
__author__ = "Kirill Klenov <horneds@gmail.com>"
__license__ = "MIT"


class JSONResponse:

    """Accept JSON data.

    Helper which makes JSON responses more easy.

    ::

        return JSONResponse(json={'json': 'here'})

    """

    def __init__(self, *, json=None, **kwargs):
        """Convert JSON data to text."""
        if json:
            kwargs.setdefault('text', ujson.dumps(json))
            kwargs.setdefault('content_type', 'application/json')
        super(JSONResponse, self).__init__(**kwargs)


class RESTNotFound(JSONResponse, HTTPNotFound):

    """Resource is not found."""

    pass


class RESTBadRequest(JSONResponse, HTTPBadRequest):

    """Request data is bad."""

    pass


class RESTForbidden(JSONResponse, HTTPForbidden):

    """Access to resource is forbidden."""

    pass


# Import Muffin-REST Elements to the root module namespace
from .api import *      # noqa
from .filters import *  # noqa
from .handlers import * # noqa

try:
    from .peewee import PWRESTHandler, PWFilter # noqa
except ImportError:
    pass



"""!!! STILL NOT IMPLEMENTED. WORK IN PROGRESS !!!."""
import asyncio
import re
from types import FunctionType

import muffin
from aiohttp import web
from muffin.urls import ParentResource


PREFIX_RE = re.compile('(/|\s)')


class ApiRoute(web.Route):

    """Support multiple routers."""

    def __init__(self, api):
        """Initialize the route."""
        self.api = api
        self.router = web.UrlDispatcher()
        super(ApiRoute, self).__init__('*', None, api.prefix_name)

    def match(self, path):
        """Check path for API."""
        if path.startswith(self.api.prefix):
            path = path[len(self.api.prefix):]
            for name, route in self.router.items():
                match_info = route.match(path)
                if match_info is not None:
                    match_info['api-name'] = name
                    return match_info
        return None

    @asyncio.coroutine
    def handler(self, request):
        """Handle request."""
        route = self.router[request.match_info['api-name']]
        response = yield from route._handler(request)
        return response

    def url(self, **kwargs):
        """Do nothing for now."""
        return self.api.prefix


class Api():

    """Bind group of resources together."""

    def __init__(self, app, prefix='/api', scheme=False):
        """Initialize the API."""
        self.app = app
        self.prefix = prefix.rstrip('/')
        self.prefix_name = PREFIX_RE.sub('.', prefix.strip('/'))
        self.handlers = {}
        self.resource = ParentResource(self.prefix, name=self.prefix_name)
        self.app.router._reg_resource(self.resource)
        if scheme:
            path = '/'.join((self.prefix, scheme.strip('/')))
            handler = muffin.Handler.from_view(self.render_scheme, 'GET')
            app.register(path)(handler)

    def register(self, *paths, methods=None, name=None):
        """Register handler to the API."""
        if isinstance(methods, str):
            methods = [methods]

        def wrapper(handler):

            if isinstance(handler, FunctionType):
                handler = muffin.Handler.from_view(handler, *(methods or ['GET']))

            if handler.name in self.handlers:
                raise muffin.MuffinException('Handler is already registered: %s' % handler.name)

            self.handlers[handler.name] = handler

            handler.connect(
                self.app, *paths, methods=methods, name=name or handler.name,
                router=self.resource.router)

            return handler

        # Support for @app.register(func)
        if len(paths) == 1 and callable(paths[0]):
            view = paths[0]
            paths = []
            return wrapper(view)

        return wrapper

    def render_scheme(self, request):
        """Render API Scheme."""
        response = {}

        for name, handler in self.handlers.items():
            scheme = {'methods': handler.methods, 'desc': handler.__doc__}
            scheme.update(getattr(handler, 'scheme', lambda: {})())
            response[name] = scheme

        return response

#  pylama:ignore=W1401,W0212



"""Support API filters."""
import operator

from marshmallow import fields, missing


class Filter:

    """Base filter class."""

    operators = {
        '$lt': operator.lt,
        '$le': operator.le,
        '$gt': operator.gt,
        '$ge': operator.ge,
        '$eq': operator.eq,
        '$ne': operator.ne,
        '$in': lambda v, c: v in c,
    }

    list_ops = '$in',

    field_cls = fields.Raw

    def __init__(self, name, fname=None, field=None):
        """Initialize filter.

        :param name: Column/Property name.
        :param fname: Filter name
        :param fields: Marshmallow Field instance

        """
        self.field = field or self.field_cls()
        self.fname = fname or name
        self.name = name

    def __repr__(self):
        """String representation."""
        return '<Filter %s>' % self.name

    def parse(self, data):
        """Parse operator and value from filter's data."""
        val = data.get(self.fname, missing)
        if not isinstance(val, dict):
            return (self.operators['$eq'], self.field.deserialize(val)),

        return tuple(
            (
                self.operators[op],
                (self.field.deserialize(val)) if op not in self.list_ops else [
                    self.field.deserialize(v) for v in val])
            for (op, val) in val.items() if op in self.operators
        )

    def apply(self, collection, ops, **kwargs):
        """Apply the filter to collection."""
        validator = lambda obj: all(op(obj, val) for (op, val) in ops)  # noqa
        return [o for o in collection if validator(o)]

    def filter(self, collection, data, **kwargs):
        """Filter given collection."""
        ops = self.parse(data)
        collection = self.apply(collection, ops, **kwargs)
        return ops, collection


class Filters:

    """Build filters for given handler."""

    FILTER_CLASS = Filter

    def __init__(self, *filters, handler=None):
        """Initialize object."""
        self.filters = tuple(
            f if isinstance(f, Filter) else self.convert(f, handler) for f in filters)

    def convert(self, args, handler=None):
        """Prepare filters."""
        name = args
        field = fname = None
        if isinstance(args, (list, tuple)):
            name, params = args
            field = params.get('field')
            fname = params.get('fname')

        if not handler or not handler.Schema or name not in handler.Schema._declared_fields:
            return self.FILTER_CLASS(name, fname=fname, field=field)

        field = field or handler.Schema._declared_fields[name]
        return self.FILTER_CLASS(name, fname=fname, field=field)

    def filter(self, data, collection, **kwargs):
        """Filter given collection."""
        if not data or self.filters is None:
            return None, collection

        filters = {}
        for f in self.filters:
            if f.fname not in data:
                continue
            ops, collection = f.filter(collection, data, **kwargs)
            filters[f.fname] = ops

        return filters, collection
